package com.blackledge.david.game;

/**
 * Represents a Deck of cards.
 *
 * @version $Revision: 1.1 $
 */
public class CardDeck {
    CardStack deck;
    private static final int suits[] = 
        new int[] {Card.CLUB, Card.HEART, Card.DIAMOND, Card.SPADE};
    /**
     * Create a Deck of cards containing the cards in the supplied stack.
     * @param cards
     */
    public CardDeck(CardStack cards) {
        deck = cards;
    }
    /**
     * Creates a new Deck of cards containing every card from
     * {@link Card#getMinValue()} to {@link Card#getMaxValue()} for all 4 suits,
     * and calls {@link #shuffle()}.
     */
    public CardDeck() {
        deck = new CardStack();
        for(int j = 0 ; j < suits.length ; ++j) {
            for(int i = Card.getMinValue() ; i <= Card.getMaxValue(); ++i) {
                deck.add(new Card(suits[j], i, false));   
            }
        }
        
        shuffle();
    }
    /**
     * perform a "natural shuffle" 3 times. cuts the deck in half and randomly
     * adds cards from the halves to the new pile
     */
    public void shuffle() {
        int shuffleRepetitions = 3;
        for(int i = 0 ; i < shuffleRepetitions ; ++i) {
            CardStack oldStack = deck;
            deck = new CardStack();
            int i1 = 0;
            int i2 = oldStack.size()/2;
            int max1 = i2;
            int max2 = oldStack.size();
            while(i1 < max1 || i2 < max2) {
                boolean half1 = Math.random() >= 0.5;
                int choice;
                if(i2 >= max2 || (half1 && i1 < max1)) {
                    choice = i1;
                    ++i1;
                } else {
                    choice = i2;
                    ++i2;
                }
                deck.add( oldStack.get( choice ) );
            }
        }
        
    }
    
    /**
     * 
     * @return The size of the deck. 
     */
    public int size() {
        return deck.size();
    }
    
    /**
     * Override for different default draw behavior.
     * By default, calls {@link #draw(int)} with a 1. 
     * @return the result of {@link #draw(int)}
     */
    public CardStack draw() {
        return draw(1);
    }
    /**
     * Draw the specified number of cards from the deck.
     * @param num the number of cards to (attempt to) draw
     * @return a CardStack containing <var>num</var> cards or {@link #size()} cards, whichever is smaller, all face up.
     */
    public CardStack draw(int num) {
        int i = num-1;
        if(deck.size() <= i) {
            i = deck.size()-1;
        }
        // returned stack is all face-up
        CardStack pickUp = deck.pickUp( i );
        if(pickUp == null) {
            return null;
        }
            
        // don't reverse it - the way decks are understood, the result is
        // stacked as-is (but face-up) on the turned-over deck and that's
        // correct.
        return pickUp;
    }
    /**
     * get the Card at the specified index, or null if the deck isn't that big.
     * @param i
     * @return the Card at that index.
     */
    public Card get(int i) {
        if(deck.size() <= i)
            return null;
        return (Card)deck.get( i );
    }
    /**
     * 
     * @return the Top Card in the deck, or null if the deck is empty.
     */
    public Card peek() {
        if(deck.size() == 0)
            return null;
        return deck.topCard();
    }
    
	public CardStack toStack() {
		return deck;
	}
}
