package com.blackledge.david.game;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;

import com.blackledge.david.game.MultiplayerGame.GameData;
import com.blackledge.david.game.MultiplayerGame.GameInfo;

public abstract class MultiplayerGameBase implements MultiplayerGame {

	private static int _nextGameID = 0;
	private static synchronized int nextGameID() {
		if(_nextGameID >= Integer.MAX_VALUE) {
			_nextGameID = 0;
		}
		return _nextGameID++;
	}
	
	protected HashMap<Integer, GameData> games = new HashMap<Integer, GameData>();
	
	public GameInfo[] getGameList() {
		GameInfo[] result = new GameInfo[games.size()];
		int i = 0;
		for(Integer id : games.keySet()) {
			GameData g = games.get(id);
			result[i] = new GameInfo();
			result[i].gameID = id;
			result[i].players = new ArrayList<PlayerState>(g.players.size());
			for(PlayerState p : g.players) {
				result[i].players.add(new PlayerState(p));
			}
			i++;
		}
		return result;
	}

	
//	public int connectPlayerClient(int gameID, GameStateListener client,
//			PlayerState playerIdentification) throws GameException {
//		if(gameID < 0) {
//			gameID = nextGameID();
//			GameData game = newGameData(gameID);
//			games.put(gameID, game);
//		}
//		
//		GameData game = games.get(gameID);
//		if(game == null) {
//			throw GameException.GAME_NOT_FOUND;
//		}
//
//		addPlayer(game, playerIdentification);
//		
////		game.gameListeners.add(client);
//		
//		// let everybody know this spectator connected (and notify client of current state)
//		notifyGameState(gameID, game.boardState, game.players);
//		
//		return gameID;
//	}
//	
//	public void connectSpectatorClient(int gameID, GameStateListener client,
//			PlayerState spectatorIdentification) throws GameException {
//		
//		//FIXME note spectators present to rest of players
//		GameData game = games.get(gameID);
//		if(game == null) {
//			throw GameException.GAME_NOT_FOUND;
//		}
//		
////		game.gameListeners.add(client);
//		
//		// let everybody know this spectator connected (and notify client of current state)
//		notifyGameState(gameID, game.boardState, game.players);
//	}
//
//	public void disconnectPlayerClient(int gameID, GameStateListener client,
//			PlayerState playerIdentification) throws GameException {
//		System.out.println("DISCONNECT PLAYER ");
//		GameData game = games.get(gameID);
//		if(game == null) {
//			throw GameException.GAME_NOT_FOUND;
//		}
//		
////		game.gameListeners.remove(client);
//		
//		removePlayer(game, playerIdentification);
//		if(game.players.size() == 0) {
//			games.remove(gameID);
//		}
//
//		notifyGameState(gameID, game.boardState, game.players);
//	}

	/**
	 * if {@link #playerAction(GameData, PlayerState, int, Object)} (passing
	 * local instance of PlayerState), then
	 * {@link #notifyGameState(int, BoardState, ArrayList)}
	 */
	public boolean playerAction(int gameID, PlayerState playerIdentification,
			int action, Object details) {
		GameData game = games.get(gameID);
		// get local instance based on playerid.
		PlayerState player = game.players.get(game.players.indexOf(playerIdentification));
		if(playerAction(game, player, action, details)) {
			notifyGameState(gameID, game.boardState, game.players);
			return true;
		}
		return false;
	}
	
	/**
	 * Get the FIRST player that has turn.
	 * @param game
	 * @return
	 */
	public PlayerState getPlayerWithTurn(int gameID) {
		return getPlayerWithTurn(getGameData(gameID));
	}
	protected PlayerState getPlayerWithTurn(GameData game) {
		for(PlayerState player : game.players) {
			if(player.hasTurn) {
				return player;
			}
		}
		return null;
	}
	
	public List<PlayerState> getPlayersWithTurn(int gameID) {
		return getPlayersWithTurn(getGameData(gameID));
	}
	protected List<PlayerState> getPlayersWithTurn(GameData game) {
		ArrayList<PlayerState> result = new ArrayList<PlayerState>();
		for(PlayerState player : game.players) {
			if(player.hasTurn) {
				result.add(player);
			}
		}
		return result;
	}
	
	private void notifyGameState(int gameID, BoardState boardState,
			ArrayList<PlayerState> players) {
		GameData gameData = games.get(gameID);
		if(gameData != null) {
			gameData.lastChanged = new Date();

			gameUpdated(gameID);
//			for(GameStateListener listener : gameData.gameListeners) {
//				listener.gameStateUpdate(boardState, players);
//			}
		}
	}
	
	private void notifyChat(int gameID, PlayerState speaker, Date timestamp, String message) {
		GameData gameData = games.get(gameID);
		if(gameData != null) {
			gameData.chat.add(new Chat(speaker, timestamp, message));
			
			chatUpdated(gameID);
//			for(ChatListener listener : gameData.chatListeners) {
//				listener.chatUpdate(gameID, speaker, timestamp, message);
//			}
		}
	}
	
	public Date getLastGameStateChange(int gameID) {
		try {
			return games.get(gameID).lastChanged;
		} catch(Exception e) {
			return new Date(0);
		}
	}
	public BoardState getBoardState(int gameID) {
		try {
			return games.get(gameID).boardState;
		} catch(Exception e) {
			return null;
		}
	}
	public ArrayList<PlayerState> getPlayers(int gameID) {
		try {
			return games.get(gameID).players;
		} catch(Exception e) {
			return new ArrayList<PlayerState>();
		}
	}
	
	public ArrayList<Chat> getChat(int gameID) {
		try {
			return games.get(gameID).chat;
		} catch(Exception e) {
			return new ArrayList<Chat>();
		}
	}

	/**
	 * Main game logic access point.  Action id and details are game/client shared understanding.
	 * @param game
	 *            GameData including current boardState that can be changed and
	 *            current player state that can be changed
	 * @param player
	 *            player performing the action (one of the instances in game.players)
	 * @param action
	 *            the action ID
	 * @param details
	 * @return true if boardState or players has changed.
	 */
	protected abstract boolean playerAction(GameData game, PlayerState player,
			int action, Object details);

	/**
	 * Initializes a new game before players are added and sets boardState to a
	 * new instance.
	 */
	public GameData newGameData(int id) {
		GameData gameData = new GameData(id);
		gameData.boardState = new BoardState(getBoardWidth(), getBoardHeight());
		return gameData;
	}
	
	public GameData getGameData(int gameID) {
		return games.get(gameID);
	}
	
	public int newGame() {
		int result = nextGameID();
		GameData gamedata = newGameData(result);
		games.put(gamedata.gameID, gamedata);

		return gamedata.gameID;
	}
	
	
	public abstract int getBoardHeight();

	public abstract int getBoardWidth();

	/**
	 * Add a player to a game and update state (e.g. start game) as appropriate.
	 * Throw full or id_in_use exceptions, modify player state or game as
	 * needed, and add player to game.players. Default impl just throws
	 * ID_IN_USE if player is already in players, otherwise adds player to
	 * game.players.
	 * 
	 * @param game
	 * @param player
	 * @throws GameException
	 */
	protected void addPlayer(GameData game, PlayerState player) throws GameException {
		// check the game isn't full of players already
//		throw GameException.GAME_FULL;
		
		// check the playerid isn't already in the game
		if(game.players.contains(player))
			throw GameException.ID_IN_USE;
		
		playerChat(game.gameID, player, "ENTERED");
		
		game.players.add(player);	
		
		if(game.players.size() == 1) {
			game.players.get(0).hasTurn = true;
		}
	}

	/**
	 * Remove a player from a game and update state (e.g. assign remaining
	 * player as winner) as appropriate. remove player from game.players.
	 * Default Impl just removes player from game.players.
	 * 
	 * @param game
	 * @param player
	 */
	protected void removePlayer(GameData game, PlayerState player) {
		game.players.remove(player);	
		playerChat(game.gameID, player, "EXITED");
	}

	/**
	 * change entries to move turn from player to next entry in game.players.
	 * @param game
	 * @param player
	 */
	protected void nextTurn(GameData game, PlayerState player) {
		int pindex = game.players.indexOf(player);
		if(game.players.size() > pindex+1) {
			++pindex;
		} else {
			pindex = 0;
		}
		game.players.get(pindex).hasTurn = true;
		player.hasTurn = false;
	}

	
//	public void connectChatListener(int gameID, ChatListener client) {
//		GameData game = games.get(gameID);
//		if(game != null) {
//			game.chatListeners.add(client);
//		}
//	}

//	public void disconnectChatListener(int gameID, ChatListener client) {
//		System.out.println("REMOVING CHAT");
//		
//		GameData game = games.get(gameID);
//		if(game != null) {
//			game.chatListeners.remove(client);
//		}
//	}

	/**
	 * Call this when Player is chatting a message... will notify other players.
	 */
	public void playerChat(int gameID, PlayerState playerIdentification,
			String message) {
		Date timestamp = new Date();
		notifyChat(gameID, playerIdentification, timestamp, message);
	}

	
	public synchronized Date getGameUpdated(int gameID) {
		return games.get(gameID).lastChanged;
	}
	public Date getPlayersUpdated(int gameID) {
		return games.get(gameID).lastChangedPlayers;
	}
	public Date getChatUpdated(int gameID) {
		return games.get(gameID).lastChangedChat;
	}
	public Date getBoardUpdated(int gameID) {
		return games.get(gameID).lastChangedBoard;
	}
	
	public synchronized BoardState getGameBoard(int gameID) {
		return games.get(gameID).boardState;
	}
	public synchronized ArrayList<PlayerState> getGamePlayers(int gameID) {
		return games.get(gameID).players;
	}
	public synchronized ArrayList<Chat> getGameChat(int gameID) {
		return games.get(gameID).chat;
	}
	
	
	

	public int joinGame(PlayerState player)
			throws GameException {
		// pass -2.. a flag that says "join an available game or start a new one"
		return joinGame(GAME_VS_ANY_PLAYER, player);
	}
	
	/**
	 * gameID of -2: join any in-progress game or create a new one.
	 * gameID of -1: create a new game and join it.
	 * gameID of -3: create a new game, and have a computer player join it.
	 */
	public int joinGame(int gameID, PlayerState playerIdentification)
					throws GameException {
		synchronized(games) {
			if(gameID == GAME_VS_ANY_PLAYER) {
				// join a game that is waiting for a player
				for(GameData game : games.values()) {
					if(game.players.size() < game.minimumPlayers) {
						gameID = game.gameID;
						break;
					}
				}
				// no games, start a new one.
			}
			boolean isAgainstComputer = false;
			if(gameID == GAME_VS_COMPUTER) {
				isAgainstComputer = true;
			}
			if(gameID < 0) {
				// start a new game
				gameID = newGame();
				gameUpdated(gameID);
			}

			GameData game = games.get(gameID);
			if(game == null) {
				throw GameException.GAME_NOT_FOUND;
			}

			if(isAgainstComputer) {
				// join a game with a computer player
				
				PlayerState computerPlayer = new PlayerState(new Date().toString(), "Computer Player");
				computerPlayer.isHuman = false;
				addPlayer(game, computerPlayer );
			}
			

			addPlayer(game, playerIdentification);
		}
		
		// let everybody know this spectator connected (and notify client of current state)
//		notifyGameState(gameID, game.boardState, game.players);
		playersUpdated(gameID);
		
		return gameID;
	}
		
	public void spectateGame(int gameID,
			PlayerState spectatorIdentification) throws GameException {
		
		//FIXME note spectators present to rest of players
		GameData game = getGameData(gameID);
		if(game == null) {
			throw GameException.GAME_NOT_FOUND;
		}
		
		
		// let everybody know this spectator connected (and notify client of current state)
//		notifyGameState(gameID, game.boardState, game.players);
		playersUpdated(gameID);
	}

	public void leaveGame(int gameID,
			PlayerState playerIdentification) throws GameException {
//		System.out.println("DISCONNECT PLAYER ");
		GameData game = getGameData(gameID);
		if(game == null) {
			throw GameException.GAME_NOT_FOUND;
		}
		removePlayer(game, playerIdentification);

//		notifyGameState(gameID, game.boardState, game.players);
		playersUpdated(gameID);
		
		// quit if that was the last human
		for(PlayerState p : game.players) {
			if(p.isHuman) {
				return;
			}
		}
		// all remaining players were computers
		games.remove(gameID);
	}
	
	/**
	 * Flag that all elements of the game should be considered changed
	 * @param gameID
	 */
	public void gameUpdated(int gameID) {
		Date timestamp = new Date();
		GameData gameData = getGameData(gameID);
		gameData.lastChanged = timestamp;
		gameData.lastChangedBoard = timestamp;
		gameData.lastChangedPlayers = timestamp;
		gameData.lastChangedChat = timestamp;
	}
	
	public void boardUpdated(int gameID) {
		Date timestamp = new Date();
		GameData gameData = getGameData(gameID);
		gameData.lastChanged = timestamp;
		gameData.lastChangedBoard = timestamp;
	}
	public void chatUpdated(int gameID) {
		Date timestamp = new Date();
		GameData gameData = getGameData(gameID);
		gameData.lastChanged = timestamp;
		gameData.lastChangedChat = timestamp;
	}
	public void playersUpdated(int gameID) {
		Date timestamp = new Date();
		GameData gameData = getGameData(gameID);
		gameData.lastChanged = timestamp;
		gameData.lastChangedPlayers = timestamp;
	}

}
