package com.blackledge.david.game.chainreaction;

import java.awt.Color;
import java.awt.Point;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;

import com.blackledge.david.game.BoardState;
import com.blackledge.david.game.GameException;
import com.blackledge.david.game.MultiplayerGameBase;
import com.blackledge.david.game.PlayerState;
import com.blackledge.david.game.MultiplayerGame.GameData;


/** http://www.atarimagazines.com/compute/issue80/chain_reaction.php */
public class ChainReactionGame extends MultiplayerGameBase {

	public static enum ACTION {
		/**
		 * add an item to this location, identified by a Point object as action
		 * details
		 */
		ADD
	}

	private HashMap<Integer, List<BoardState>> _reaction = new HashMap<Integer, List<BoardState>>();
	
	@Override
	protected boolean playerAction(GameData game, PlayerState player,
			int action, Object details) {
		if(action < 0 || action >= ACTION.values().length) {
			// illegal action
			return false;
		}
		
		switch(ACTION.values()[action]) {
		case ADD:
			Point location = (Point) details;
			if (location == null || location.x < 0 || location.y < 0
					|| location.x >= getBoardWidth()
					|| location.y >= getBoardHeight()) {
				return false;
			}
			
			BoardState state = getBoardState(game.gameID);
			if(isWin((ChainReactionGameData) game)) {
				return false;
			}
			int[] players = getPlayerValues(state.get(location.x, location.y));
			int playerIndex = game.players.indexOf(player);
			
			// can only place in blank or already-owned space
			for(int i = 0 ; i < players.length ; ++i) {
				if(i != playerIndex && players[i] > 0) {
					return false;
				}
			}
			
			players[playerIndex]+=1;
			state.set(location.x, location.y, getBoardValue(players));
			
			react((ChainReactionGameData) game, state, playerIndex);
			
			return true;
		}
		
		return false;
	}
	
	public boolean isComputerPlayerAvailable() {
		// FIXME add a computer player
		return false;
	}
	
	@Override
	public GameData newGameData(int id) {
//		GameData data = super.newGameData(id);
		GameData data = new ChainReactionGameData(id);
		data.boardState = new BoardState(getBoardWidth(), getBoardHeight());

		// init the board state
		BoardState state = data.boardState;
		for(int x = 0 ; x < state.width ; ++x) {
			for(int y = 0 ; y < state.height ; ++y) {
				state.set(x, y, 0);
			}
		}
		
		return data;
	}
	
	private Color[] colors = new Color[]{Color.blue.brighter(), Color.red, Color.green, Color.cyan};

	@Override
	protected void addPlayer(GameData game, PlayerState player)
			throws GameException {
		synchronized(colors) {
			// player may not enter with a color, it must be assigned
			player.color = null;

			// assign colors to all players
			ArrayList<Color> c = new ArrayList<Color>();
			c.addAll(Arrays.asList(colors));
			// don't use any previous players' colors
			for(PlayerState p : game.players) {
				c.remove(p.color);
			}
			if(player.color == null) {
				player.color = c.remove(0);
			}
		}
		
		super.addPlayer(game, player);
	}
	
	private void react(ChainReactionGameData game, BoardState state, int player) {
		if(_reaction.get(game.gameID) == null) {
			_reaction.put(game.gameID, new ArrayList<BoardState>());
		}
		_reaction.get(game.gameID).clear();
		// include starting (possibly only) state:
		_reaction.get(game.gameID).add(new BoardState(state));
		
		boolean reacting = true;
		int sanityCheck = 150;
		while(reacting) {
			if(--sanityCheck <= 0) {
				break;
			}
			if(isWin(game)) {
				break;
			}
			
			reacting = false;
			ArrayList<Point> explosions = new ArrayList<Point>();
			// record updates in a copy so we don't... have a chain reaction
			BoardState newState = new BoardState(state);
			for(int x = 0 ; x < state.width ; ++x) {
				for(int y = 0 ; y < state.height ; ++y) {
					int rawval = state.get(x, y);
					if(rawval < 0) {
						//TODO this is really an error state.  Everything should be 0 or above
						continue;
					}
					int max = 2;
					if(x > 0 && x < state.width-1) {
						++max;
					}
					if(y > 0 && y < state.height-1) {
						++max;
					}
					
					int[] pvals = getPlayerValues(rawval);
					
					int val = 0;
					for(int v : pvals) val += v;
					
					if(val >= max) {
						// this exploded, we'll need to do at least one more pass.
						reacting = true;
						explosions.add(new Point(x,y));
						// remove from this spot... any remaining now belong to player
						int[] players = getPlayerValues(0);
						players[player] = val-max;
						newState.set(x,y,getBoardValue(players));
						// and add to adjacent spots in player's slot
						if(x > 0) {
							reactInto(player, newState, x-1, y);
						}
						if(y > 0) {
							reactInto(player, newState, x, y-1);
						}
						if(x < state.width-1) {
							reactInto(player, newState, x+1, y);
						}
						if(y < state.height-1) {
							reactInto(player, newState, x, y+1);
						}
					}
					
				}
			}

			// record explosions in a special BoardState instance
			BoardState explosionState = new BoardState(newState.width, newState.height);
			for(Point exp : explosions) {
				explosionState.set(exp.x, exp.y, -100);
			}
			_reaction.get(game.gameID).add(explosionState);
			// record post-explosion state
			_reaction.get(game.gameID).add(newState);
			
			// newState is now the true state for the next pass
			for (int x = 0; x < state.width; ++x) {
				for (int y = 0; y < state.height; ++y) {
					state.set(x, y, newState.get(x, y));
				}
			}
			
			if(reacting) {
				// can't win until there has been at least one explosion
				((ChainReactionGameData)game).exploded = true;
			}
		}
	}
	
	private static class ChainReactionGameData extends GameData {
		public ChainReactionGameData(int gameID) {
			super(gameID);
		}

		/** whether at least one reaction has taken place */
		boolean exploded = false;
	}

	private boolean isWin(ChainReactionGameData game) {
		// can't win until there has been at least one explosion
		if(!game.exploded)
			return false;
		BoardState state = game.boardState;
		int winner = -1;
		for (int x = 0; x < state.width; ++x) {
			for (int y = 0; y < state.height; ++y) {
				int[] players = getPlayerValues(state.get(x, y));
				for(int i = 0 ; i < players.length ; ++i) {
					if(players[i] > 0) {
						if(winner < 0) {
							winner = i;
						} else {
							if(i != winner) {
								winner = -1;
								return false;
							}
						}
					}
				}
			}
		}
		return winner > -1;
	}

	private void reactInto(int player, BoardState state,
			int px, int py) {
		int[] players;
		players = getPlayerValues(state.get(px, py));
		int pval = 0;
		for(int v : players) pval += v;
		players = getPlayerValues(0);
		players[player] = pval + 1;
		++players[player];
		state.set(px, py, getBoardValue(players));
	}

	public int[] getPlayerValues(int rawval) {
		int numplayers = 2;//FIXME need to allow for other numbers, not hardcoded
		int modifier = 100;
		int result[] = new int[numplayers];
		for(int i = numplayers-1 ; i >= 0 ; --i) {
			result[i] = rawval % modifier;
			rawval = (rawval - result[i]) / modifier;
		}
		return result;
	}

	private int getBoardValue(int[] players) {
		int val = 0;
		int modifier = 0;
		for (int p : players) {
			val = val * modifier + p;
			modifier += 100;
		}
		return val;
	}

	// represent explosion points, new board state, explosion points, new board state, etc. until final state.
	public List<BoardState> getReaction(int gameID) {
		return _reaction.get(gameID);
	}

	@Override
	public int getBoardHeight() {
		return 5;
	}

	@Override
	public int getBoardWidth() {
		return 6;
	}
}
