package com.blackledge.david.game.cribbage;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;
import java.util.List;

import com.blackledge.david.game.BoardState;
import com.blackledge.david.game.Card;
import com.blackledge.david.game.CardDeck;
import com.blackledge.david.game.CardStack;
import com.blackledge.david.game.GameException;
import com.blackledge.david.game.MultiplayerGameBase;
import com.blackledge.david.game.PlayerState;

public class CribbageGame extends MultiplayerGameBase {

	@Override
	public boolean playerAction(GameData game, PlayerState player,
			int action, Object details) {
		boolean result = false;
		if(player.hasTurn) {
			int gameID = game.gameID;
			// note last player action
			game.lastPlayerAction = new Date();
			
			ACTION act;
			try {
				act = ACTION.values()[action];
				switch(act) {
				case DEAL:
					result = actionDeal(gameID, player);
					break;
				case DISCARD:
					result = actionDiscard(gameID, player, details);
					break;
				case PLAY:
					result = actionPlay(gameID, player, details);
					break;
				case GO:
					result = actionGo(gameID, player);
					break;
				}
			} catch (Exception e) {
				result = false;
			}
		}
		
		return result;
	}

	@Override
	public Date getBoardUpdated(int gameID) {
		allowComputerPlayerAction(gameID);
		Date result = super.getBoardUpdated(gameID);
		return result;
	}

	@Override
	public synchronized Date getGameUpdated(int gameID) {
		allowComputerPlayerAction(gameID);
		Date result = super.getGameUpdated(gameID);
		return result;
	}
	
	@Override
	public Date getPlayersUpdated(int gameID) {
		allowComputerPlayerAction(gameID);
		Date result = super.getPlayersUpdated(gameID);
		return result;
	}
	
	@Override
	public Date getLastGameStateChange(int gameID) {
		allowComputerPlayerAction(gameID);
		Date result = super.getLastGameStateChange(gameID);
		return result;
	}
	
	@Override
	public Date getChatUpdated(int gameID) {
		allowComputerPlayerAction(gameID);
		Date result = super.getChatUpdated(gameID);
		return result;
	}
	
	public boolean isComputerPlayerAvailable() {
		return true;
	}
	
	private long _computerPauseMillis = 1750;
	
	private void allowComputerPlayerAction(int gameID) {
		// this is our chance to drive AI player actions
		// assumption is getLastGameStateChange is regularly polled by human player interfaces.
		Date now = new Date();
		GameData game = getGameData(gameID);
		if (game.lastPlayerAction == null
				|| game.lastPlayerAction.getTime() < (now.getTime() - _computerPauseMillis)) {
			// if computer player has turn, game.lastPlayerAction will be updated.
			doComputerPlayerTurns(game);
		}
	}

	private void doComputerPlayerTurns(GameData game) {
		for (PlayerState computerPlayer : game.players) {
			if (!computerPlayer.isHuman && computerPlayer.hasTurn) {
				computerPlayerAction((CribbageGameData) game, computerPlayer);
			}
		}
	}
	
	private int waitingForHumans;

	private void computerPlayerAction(CribbageGameData game, PlayerState computerPlayer) {
		if(computerPlayer.isHuman)
			return;
		if(computerPlayer.hasTurn) {
			switch(game.gameState) {
			case DISCARDS:
				waitingForHumans = game.players.size()-1;
				//TODO intelligent choice (discard good cards if I'm dealer, bad cards if not)

				// figure out the highest scoring hand to keep.
				List<Integer> handchoice = null;
				int topHandScore = 0;
				List<Integer> playerHand = getPlayerHand(game.gameID, computerPlayer);
				if(playerHand.size() > 4) {
					List<List<Integer>> hands = getSubsets(playerHand, 4);
					for(List<Integer>hand : hands) {
						hand.add(0, 1);// random Ace as turn-up card for score tests
						List<CribbageScore> scores = getHandScore(hand, false);
						int score = 0;
						for (CribbageScore s : scores) {
							score += s.getValue();
						}
						if (score > topHandScore) {
							topHandScore = score;
							handchoice = hand;
						}
					}
					// discard one of the remaining cards
					List<Integer> discards = new ArrayList<Integer>(playerHand);
					discards.removeAll(handchoice);
					int discardchoice = 0;
					if(discards.size() > 0) {
						discardchoice = playerHand.indexOf(discards.get(0));
					}

					playerAction(game, computerPlayer, ACTION.DISCARD.ordinal(), new Integer(discardchoice));
				}
				break;
			case GO_GOING:
			case PLAY:
				waitingForHumans = game.players.size()-1;
				//TODO could be defensive in PLAY state and aggressive in GOING state.
				
				// figure out the highest scoring card to play, but avoid making a 21.
				int playchoice = -1;
				int validchoice = -1;
				int topScore = 0;
				CardStack hand = getCardStack(game.gameID, getPlayerHandRow(computerPlayer));
				for (int i = 0; i < hand.size(); ++i) {
					Card card = (Card) hand.get(i);
					if (card.isFaceUp()) {
						// check the score for each playable card, and remember the top one.
						List<Integer> play = getCards(game.gameID, PLAY_ROW);
						play.add(getCardRef(card));
						
						// is it legal?
						int count = getCount(play) - game.countOffset;
						if(count <= 31) {
							validchoice = i;
							// what would it score?
							List<CribbageScore> scores = getPlayScore(game, play);
							int score = 0;
							for (CribbageScore s : scores) {
								score += s.getValue();
							}
							if (score > topScore
									// avoid 21, though
									|| (score == topScore && count != 21)) {
								topScore = score;
								playchoice = i;
							}
						}
					}
				}
				// could have no "good" choices, but have to choose SOMETHING
				if(playchoice < 0) playchoice = validchoice;
				playerAction(game, computerPlayer, ACTION.PLAY.ordinal(), new Integer(playchoice));
				break;
			case GO_NEXT:
				waitingForHumans = game.players.size()-1;
				// say "Go."
				playerAction(game, computerPlayer, ACTION.GO.ordinal(), new Integer(0));
				break;
				
			case GAME_WAITING:
				waitingForHumans = game.players.size();
				break;
			case GAME_WIN:
				if(waitingForHumans <= 0) {
					// TODO start new game
					waitingForHumans = game.players.size()-1;
				} else {
					--waitingForHumans;
				}
				break;
			case ROUND_ENDED:
			case ROUND_BEGIN:
			case TURN_UP:
				if(waitingForHumans <= 0) {
					playerAction(game, computerPlayer, ACTION.DEAL.ordinal(), new Integer(0));
					waitingForHumans = game.players.size()-1;
				} else {
					--waitingForHumans;
				}
				break;
			}
		}
	}

	@Override
	public int getBoardHeight() {
		// 1: deck
		// 1: played cards
		// 1: crib
		// 4: up to 4 hands
		
		return 7;
	}

	@Override
	public int getBoardWidth() {
		// 52 for the deck
		// play area needs at most 14: 4 a + 4 2 + 4 3 + 2 4  > 31
		// crib needs at most 4
		// hands need at most 6
		return 52;
	}

	private Color[] colors = new Color[]{Color.blue.brighter(), Color.red, Color.green, Color.cyan};
	
	@Override
	protected void addPlayer(GameData gamed, PlayerState player)
			throws GameException {
		CribbageGameData game = (CribbageGameData) gamed;
		// IF SAVED GAME, only allow previous player to come in.
		// FIXME assign a color which also indicates hand number
		synchronized(colors) {
			// player may not enter with a color, it must be assigned
			player.color = null;

			// assign colors to all players
			ArrayList<Color> c = new ArrayList<Color>();
			c.addAll(Arrays.asList(colors));
			// don't use any previous players' colors
			for(PlayerState p : game.players) {
				c.remove(p.color);
			}
			if(player.color == null) {
				player.color = c.remove(0);
			}
		}
		// zero out the score
		player.score = 0;
			
		try {
			super.addPlayer(game, player);
		} catch(GameException e) {
			// if super impl throws exception, undo color assignment!

			// Actually, this happens automatically... player was never
			// added to game, so his color is still available to assign to
			// next player.
			throw e;
		}
		// notify state, board, and player changes
		notifyGameChanged(game, false, false, true);
		
		// if we got this far, if enough players are present, one was given a turn, but that's not how we play.
		if(game.players.size() >= game.minimumPlayers) {
			switch(game.gameState) {
			case GAME_WAITING:
				// our usual case...START THE GAME!
				startGame(game);
				break;
				//FIXME what about other cases? for now they "can't happen" unless somebody left and came back
			}
		}
	}
	
	@Override
	protected void removePlayer(GameData game, PlayerState player) {
		// FIXME If game in progress, SAVE GAME or forfeit game
		// FIXME if game not in progress we'll just wait for somebody else.
		
		Color removed = player.color;
		// remove player's info
		super.removePlayer(game, player);
	}

	
	public static class CribbageCard extends Card {
		public CribbageCard(int suit, int value, boolean faceUp) {
			super(suit, value, faceUp);
		}

		/** no greater than 10 */
		public int getCribbageCountValue() {
			if(getValue() > 10) return 10;
			return getValue();
		}
	}
	

		private static class CribbageScore {
			private int _value;
			private String _name;
			private Card[] _cards;

			public CribbageScore(String name, int value) {
				this(name, value, (Card[])null);
			}
			
			public CribbageScore(String name, int value, Card[] cards) {
				_name = name;
				_value = value;
				_cards = cards;
			}

			public CribbageScore(String name, int value, List<Integer> cards) {
				this(name, value);
				_cards = new CribbageCard[cards.size()];
				for(int i = 0 ; i < cards.size(); ++i ) {
					_cards[i] = getCard(cards.get(i));
				}
			}
			
			public long getValue() {
				return _value;
			}

			public String getName() {
				return _name;
			}
			
			@Override
			public String toString() {
				if(_cards != null && _cards.length > 0) {
					return getName()+" "+Arrays.asList(_cards);
				} else {
					return getName();
				}
			}

		}

		public static final int DECK_ROW = 0;
		public static final int PLAY_ROW = 1;
		public static final int CRIB_ROW = 2;
		public static final int PLAYER1_ROW = 3;
		public static final int PLAYER2_ROW = 4;
//		public static final int PLAYER3_ROW = 5;
//		public static final int PLAYER2_ROW = 6;
		
		/** number of cards in a hand when discard phase is complete */
		public static final int PLAY_HAND_SIZE = 4;
		/** number of cards in a hand when round dealing is complete */
		public static final int DEALT_HAND_SIZE = 6;
		
		
		public static enum GAMESTATE {GAME_WAITING, ROUND_BEGIN, DISCARDS, TURN_UP, PLAY, GO_NEXT, GO_GOING, ROUND_ENDED, GAME_WIN };

		public static class CribbageGameData extends GameData {
			public CribbageGameData(int gameID) {
				super(gameID);
			}
			public GAMESTATE gameState = GAMESTATE.GAME_WAITING;
			public int dealer = 0;
			/** amount to subtract due to previous 31-counts */
			public int countOffset = 0;
		
		}
		
		
		public static enum ACTION {
			DEAL,
			/** details should be an Integer index in the hand of the card getting discarded */
			DISCARD, 
			/** details should be an Integer index in the hand of the card getting played */
			PLAY, 
			GO
			};

		// new round: deal new hands, begin discards
		// discards: until player has 4, until all players have 4, begin play
		// play: if turn, play one card, do scoring, begin round win, go, or next turn
		// go: if turn, play one card, do scoring, begin round win or go again
		// round win: pause to show hand scoring, do final scoring, pause to show score, begin win or new round
		// win: show win... some way to start a new game
		private boolean actionDeal(int gameID, PlayerState player) {
			CribbageGameData game = (CribbageGameData) games.get(gameID);
			if(game.gameState == GAMESTATE.ROUND_BEGIN) {
				dealGame(game);
				return true;
			} else {
				return false;
			}
		}
		
		private boolean actionGo(int gameID, PlayerState player) {
//			List<Integer> playerHand = getPlayerHand(gameID, player);
			CribbageGameData game = (CribbageGameData) games.get(gameID);
			if (!(game.gameState.equals(GAMESTATE.GO_NEXT))) {
				return false;
			}
			// chat "GO", give turn back and change game state to GO.
			playerChat(gameID, player, "Go! (I can't play any more)");
			nextTurn(game, player);

			// The Go point goes to the last card played of the sequence, not now. (important for 3+ players)
			// FIXME on sequence start, only players that still have cards are playing (2 players: no go is stated, though point for last card would still happen)
			
			PlayerState otherPlayer = getPlayerWithTurn(game);
			if(isUnableToPlay(game, otherPlayer)) {
				playerChat(gameID, otherPlayer, "I can't play, either");
				// so the last played hand deserved the "go" point
				//FIXME figure out who was LAST and deserves the points (check owner of last card in play area).
				int count = getCount(getCards(game.gameID, PLAY_ROW)) - game.countOffset;
				if (addGoPoints(otherPlayer, game, count)) {
					doGameWon(game);
				} else {
					if(isRoundOver(game)) {
						doRoundOver(game);
					} else {
						doSequenceOver(otherPlayer, game);
					}
				}
			} else {
				game.gameState = GAMESTATE.GO_GOING;
			}
			// notify state, board, and player changes
			notifyGameChanged(game, true, false, true);
			return true;
		}

		private boolean actionDiscard(int gameID, PlayerState player,
				Object details) {
			List<Integer> playerHand = getPlayerHand(gameID, player);
			CribbageGameData game = (CribbageGameData) games.get(gameID);
			if(!game.gameState.equals(GAMESTATE.DISCARDS)) {
				return false;
			}
			if(playerHand.size() <= PLAY_HAND_SIZE) {
				return false;
			}

			// move specific card index (from "details") onto crib (face down)
			int card = (Integer) details;
			// actually remove card from hand since it's discarded (versus how PLAY is handled)
			if(card >= playerHand.size()) return false;
			Integer cardRef = playerHand.remove(card);
			setPlayerHand(gameID, player, playerHand);
			
			List<Integer> crib = getCards(gameID, CRIB_ROW);
			crib.add(cardRef);
			setCards(gameID, CRIB_ROW, crib);
			
			// check if all players are done discarding, if so, change
			// game state to PLAY and assign first turn
			boolean stillDiscarding = false;
			for(int i = 0 ; i < game.players.size() ; ++i) {
				stillDiscarding |= (getPlayerHand(gameID, game.players.get(i)).size() > PLAY_HAND_SIZE);
			}
			if(! stillDiscarding) {
				startRound(game);
			} else {
				// notify state, board, and player changes
//				notifyGameChanged(game, false, true, false);
				notifyGameChanged(game, true, true, true);//TODO just claim all so it shows up!!
			}

			return true;
		}

		private boolean actionPlay(int gameID, PlayerState player,
				Object details) {
			List<Integer> playerHand = getPlayerHand(gameID, player);
			CribbageGameData game = (CribbageGameData) games.get(gameID);
			if (!(game.gameState.equals(GAMESTATE.PLAY)
					|| game.gameState.equals(GAMESTATE.GO_GOING))) {
				return false;
			}
			if(getRemainingCardCount(playerHand) < 1) {
				return false;
			}
			
			// move specific card index (from "details") onto playing area (face up)
			int cardindex = (Integer) details;
			if(cardindex >= playerHand.size()) {
				return false;
			}
			Integer cardRef = playerHand.get(cardindex);
			if(cardRef < 0) {
				return false;
			}
			// is it a legal play?
			List<Integer> play = getCards(gameID, PLAY_ROW);
			int count = getCount(play) - game.countOffset;
			if (count + getCard(cardRef).getCribbageCountValue() > 31) {
				return false;
			}
			
			// keep track of card but it's no longer available
			playerHand.set(cardindex, -cardRef);
			setPlayerHand(gameID, player, playerHand);
			
			play.add(cardRef);
			setCards(gameID, PLAY_ROW, play);
			// update count before we announce it
			count = getCount(play) - game.countOffset;
			
			// announce cumulative count
			playerChat(gameID, player, String.valueOf(count));

			if(getRemainingCardCount(playerHand) == 0) {
				
//				//FIXME hack to get other player's index
//				PlayerState otherPlayer = game.players.get(1-game.players.indexOf(player));
//				List<Integer> otherHand = getPlayerHand(gameID, otherPlayer);
//				if(getRemainingCardCount(otherHand) == 0) {
//					// hand is over... deal again.
//				}
			}
			
			// calculate score for changed playing area, "chat" each scoring item, and add score to current player's total
			List<CribbageScore> scores = getPlayScore(game, play);
			for(CribbageScore score : scores) {
				playerChat(gameID, player, score.toString());
				player.score += score.getValue();
			}
			
			// Check for win, change game state, etc.
								
			if(isGameWon(game)) {
				doGameWon(game);
			} else if(isRoundOver(game)) {
//				if (game.gameState == GAMESTATE.GO_GOING) {
					if (addGoPoints(player, game, count)) {
						doGameWon(game);
						return true;
					}
//				}
				doRoundOver(game);
			} else if (getCount(getCards(game.gameID, PLAY_ROW))-game.countOffset == 31) {
				// player just played the 31 count... time to reset and keep going
				// (this acts as last card, "eligible for go points" although
				//  not really possible, and "isGameWon" was already checked)
				if (addGoPoints(player, game, count)) {
					doGameWon(game);
					return true;
				}
				doSequenceOver(player, game);
			} else if (game.gameState == GAMESTATE.GO_GOING) {
				// if game state is "GO" don't change turn because we "own" turn
				if(isUnableToPlay(game, player)) {
					if (addGoPoints(player, game, count)) {
						doGameWon(game);
						return true;
					}
					// if this were due to round-over, that was already handled in earlier else-if
//					playerChat(gameID, player, "I can't play any more."); // "one for the go"/31 accounts for this.
					doSequenceOver(player, game);
				}
			} else {
				// next turn
				// Check if next player is a new "go" state
					// ... setup where only action available to next player is "GO" 
				nextTurn(game, player);
				PlayerState nextPlayer = getPlayerWithTurn(game);
				if(isUnableToPlay(game, nextPlayer)) {
					// set game state to GO so user can "do the go"
					game.gameState = GAMESTATE.GO_NEXT; 
				} else {
					// good to go...turn already changed
				}
			}
			
			// notify game board, player (score) change, and maybe game state
			// notify state, board, and player changes
			notifyGameChanged(game, true/*TODO ?*/, true, true);
			return true;
		}

		// return true if adding go points causes a win
		private boolean addGoPoints(PlayerState player, CribbageGameData game,
				int count) {
			// if player didn't just score 31, then they get a point for playing the last card.
			if(count < 31) {
				player.score += 1;
				playerChat(game.gameID, player, "One for the Go (last card).");
			}
			return isGameWon(game);
		}

		private void doSequenceOver(PlayerState player, CribbageGameData game) {
			game.countOffset = getCount(getCards(game.gameID, PLAY_ROW));
			game.gameState = GAMESTATE.PLAY;
			playerChat(game.gameID, player, "New Count. Your turn.");
			nextTurn(game, player);
			//HOWEVER new turn player might be out of cards already!
			PlayerState otherPlayer = getPlayerWithTurn(game);
			if(isUnableToPlay(game, otherPlayer)) {
				game.gameState = GAMESTATE.GO_NEXT;
			}
		}

		private void doRoundOver(CribbageGameData game) {
			game.gameState = GAMESTATE.ROUND_ENDED;
			playerChat(game.gameID, game.players.get(game.dealer), "Time to peg the hands and crib");
			game.countOffset = 0;
			
			// count up pone's hand, check for a winner
			//FIXME hack to get pone's index
			PlayerState pone = game.players.get(1-game.dealer);
			List<CribbageScore> poneScore = getHandScore(game, pone);
			int points = 0;
			if(poneScore.size() == 0) {
				playerChat(game.gameID, pone, "I have nothing");
			} else {
				for (CribbageScore score : poneScore) {
					playerChat(game.gameID, pone, score.toString());
					points += score.getValue();
				}
				playerChat(game.gameID, pone, "I pegged "+points+" points");
				pone.score += points;
				points = 0;
			}
			
			// check for win
			if(isGameWon(game)) {
				doGameWon(game);
			} else {				
				// then count up dealer's hand, and dealer's crib, check for a winner
				PlayerState dealer = game.players.get(game.dealer);
				List<CribbageScore> dealerScore = getHandScore(game, dealer);
				points = 0;
				if(dealerScore.size() == 0) {
					playerChat(game.gameID, dealer, "I have nothing in my hand");
				} else {
					for (CribbageScore score : dealerScore) {
						playerChat(game.gameID, dealer, score.toString());
						points += score.getValue();
					}
				}
				// crib
				List<CribbageScore> cribScore = getHandScore(game, dealer, true);
				if(cribScore.size() == 0) {
					playerChat(game.gameID, dealer, "I have nothing in my crib");
				} else {
					for (CribbageScore score : cribScore) {
						playerChat(game.gameID, dealer, "Crib: " + score.toString());
						points += score.getValue();
					}
				}
				if(points > 0) {
					playerChat(game.gameID, dealer, "I pegged "+points+" points");
					dealer.score += points;
					points = 0;
				}
				// check for win
				if(isGameWon(game)) {
					doGameWon(game);
				} else {	
					// if both players are out of cards, set dealer and deal
					if(pone.score < dealer.score) {
						playerChat(game.gameID, dealer, "I'm winning so far");
					} else if(pone.score > dealer.score) {
						playerChat(game.gameID, dealer, "You're winning so far");
					} else {
						playerChat(game.gameID, dealer, "We're tied!");
					}
					
					// start the next hand.
					startGame(game);
				}
			}
		}

		private void doGameWon(CribbageGameData game) {
			game.gameState = GAMESTATE.GAME_WIN;
			for(PlayerState p : game.players) {
				p.hasTurn = false;
				if(p.score >= 121) {
					playerChat(game.gameID, p, "I Win!");
				}
			}
			notifyGameChanged(game, true, true/*?*/, true);
		}

		/** count the number of cards in hand that are still usable (non-negative) */
		private int getRemainingCardCount(List<Integer> playerHand) {
			int count = 0;
			for (int card : playerHand) {
				if (card > 0) {
					++count;
				}
			}
			return count;
		}

		public boolean isGameWon(CribbageGameData game) {
			return getGameWinner(game) != null;
		}
		
		public PlayerState getGameWinner(CribbageGameData game) {
			for(PlayerState player : game.players) {
				if(player.score >= 121) {
					return player;
				}
			}
			return null;
		}

		private boolean isRoundOver(CribbageGameData game) {
			for(PlayerState player : game.players) {
				if(getRemainingCardCount(getPlayerHand(game.gameID, player)) != 0) {
//				if(! isUnableToPlay(game, player)) {
					// somebody can play, still.
					return false;
				}
			}
			return true;
		}

		private boolean isUnableToPlay(CribbageGameData game,
				PlayerState player) {
			List<Integer> play = getCards(game.gameID, PLAY_ROW);
			int count = getCount(play) - game.countOffset;
			if(count >= 31) return true;
			
			List<Integer> handi = getPlayerHand(game.gameID, player);
			ArrayList<CribbageCard> hand = getCards(handi);
			for (CribbageCard card : hand) {
				if (card.getCribbageCountValue() + count <= 31) {
					// able to play.
					return false;
				}
			}
			return true;
		}

		private void startRound(CribbageGameData game) {
			
//			List<Integer> lastdeck = getCards(game.gameID, DECK_ROW);
//			// break down previous round
//			List<Integer> lastplay = getCards(game.gameID, PLAY_ROW);
//			lastdeck.addAll(lastplay);
//			lastplay.clear();
//			setCards(game.gameID, DECK_ROW, lastdeck);
//			setCards(game.gameID, PLAY_ROW, lastplay);
			
			game.gameState = GAMESTATE.TURN_UP;
			// when we begin, turn-up turn belongs to dealer, but we'll change that momentarily
			PlayerState dealer = game.players.get(game.dealer);
			for (int i = 0; i < game.players.size(); ++i) {
				game.players.get(i).hasTurn = false;
			}
			dealer.hasTurn = true;
			
			
			// turn over the first card and give any score to the dealer
			List<Integer> deck = getCards(game.gameID, DECK_ROW);
			int card = deck.remove(0);
			// turn it face up
			if(card < 0) {
				card = -card;
			}
			List<Integer> play = getCards(game.gameID, PLAY_ROW);
			play.add(card);
			setCards(game.gameID, DECK_ROW, deck);
			setCards(game.gameID, PLAY_ROW, play);

			// turn-up doesn't count towards play, only hands.
			int count = getCount(play);
			game.countOffset = count;
			
			// a single card that is a Jack (during TURN_UP) gives 2 points
			if(getCard(card).getValue() == Card.JACK) {
				CribbageScore score = new CribbageScore("Two for His Heels (Nibs)", 2, new Card[] {getCard(card)});
				playerChat(game.gameID, dealer, score.toString());
				dealer.score += score.getValue();
			}

			// in case rules allow some other score here (although they don't)
			List<CribbageScore> scores = getPlayScore(game, play);
			for(CribbageScore score : scores) {
				playerChat(game.gameID, dealer, score.toString());
				dealer.score += score.getValue();
			}
			
			// it's possible for the dealer to win on the above points
			if(isGameWon(game)) {
				doGameWon(game);
				return;
			}

			// now begin gameplay with non-dealer's turn
			// (this is a subset of doSequenceOver)
			game.gameState = GAMESTATE.PLAY;
			nextTurn(game, dealer);
			
			// notify state, board, and player changes
			notifyGameChanged(game, true, true, true);
		}

		/** @deprecated */
		public static CribbageCard getCard(Integer cardRef) {
//			boolean faceup = true;
			if(cardRef < 0) {
//				faceup = false;
				cardRef = -cardRef;
			}
			int val = cardRef % 100;
			int suit = (cardRef-val)/100;
//			return new CribbageCard(suit, val, faceup);
			return new CribbageCard(suit, val, true);
		}
		
		/** @deprecated */
		private static Integer getCardRef(Card card) {
			int val = card.getValue();
			int suit = card.getSuit();
			
			int ref = val+(100*suit);
			return ref;
		}
		
		private List<CribbageScore> getPlayScore(CribbageGameData game, List<Integer> cards) {

			// In the playing phase of Cribbage, the players take it in turns to
			// lay down a card, trying to make the running total equal to
			// certain values. The non-dealer plays first and states the value
			// of her card (for example, "ten" for a Jack). Court cards count
			// ten (together with the face 10 they are known as the 'ten-cards',
			// or 'tenth cards'). Ace counts one.
			//
			// 15 and 31
			// The dealer then plays a card, the value of which is added to the
			// current running total. The player who makes the total exactly 15
			// scores two points ("fifteen-two"). Two points are also awarded
			// for making 31. Additionally, you score a point if your opponent
			// cannot play without going over 31 ("one for the go", or just
			// "one for go"). You must play if you can (reneging is against the
			// rules).
			//
			//
			
			// make a copy
			cards = new ArrayList<Integer>(cards);
			
			// don't include any cards that came before the last "go"
			int goStart = cards.size();
			while(goStart > 0 && getCount(cards.subList(0,  goStart)) > game.countOffset) {
				--goStart;
			}
			//TODO could get here with getCount > game.countOffset, but that's something VERY wrong with board.
			cards = cards.subList(goStart, cards.size());
			
			List<CribbageScore> result = getPlayScore(cards, game.gameState == GAMESTATE.GO_GOING);
			
			return result;
		}

		public static List<CribbageScore> getPlayScore(List<Integer> cards, boolean isGo) {
			List<CribbageScore> result = new ArrayList<CribbageScore>();

			int count = getCount(cards);
			if(count == 15) {
				result.add(new CribbageScore("Fifteen-Two", 2, cards));
			}
			else if(count == 31) {
//				if(isGo) { // go is now only scored at the end if 31 was not scored.
//					result.add(new CribbageScore("Thirty-one (after Go)", 1, cards));
//				} else {
					result.add(new CribbageScore("Thirty-one", 2, cards));
//				}
			}
			
			// Pairs
			// If your card is the same rank as the last card played, you score
			// two for a pair. If your opponent plays a third card of the same
			// rank, he scores 6 for a "pair royal" (three of a kind). Four of a
			// kind scores 12 ("double pair royal").
			
			// check "of a kinds" at the end of the play area
			if (cards.size() > 1) {
				CribbageCard card1 = getCard(cards.get(cards.size() - 1));
				int kind = card1.getValue();
				
				CribbageCard card2 = getCard(cards.get(cards.size() - 2));
				if(kind == card2.getValue()) {
					// it's a least a pair
					CribbageScore score = new CribbageScore("Pair", 2, new Card[] { card1, card2 });
					if (cards.size() > 2) {
						CribbageCard card3 = getCard(cards.get(cards.size() - 3));
						if (kind == card3.getValue()) {
							// it's at least royal
							score = new CribbageScore("Pair Royal (3-of-a-kind)", 6, new Card[] { card1, card2, card3 });
							if (cards.size() > 3 ) {
								CribbageCard card4 = getCard(cards.get(cards.size() - 4));
								if (kind == card4.getValue()) {
									// it's double royal!
									score = new CribbageScore("Double Pair Royal (4-of-a-kind)", 12, new Card[] { card1, card2, card3, card4 });
								}
							}
						}
					}
					result.add(score);
				}
			}
					
			
			// Runs
			// If the last 3 cards played form a sequence, the player making the
			// sequence scores 3 for a "run". For example, 3-4-5 makes a run of
			// 3 and so scores 3 for the player laying down the 5. If the
			// opponent then plays a 6 (or a 2) to extend the sequence to 4
			// cards, she scores 4, and so on as long as the sequence is
			// unbroken.
			//
			// Sequence do not have to be in order. For example, if the play
			// goes 7-9-6, you can then play an 8 to score 4 for a run of 4.
			if(cards.size() >= 3) {
				ArrayList<CribbageCard> run = getCards(cards);
				int score = 0;
				Card[] scorecards = null;
				// get increasingly smaller sets of cards, sort the subset and see if it's a run
				for(int i = 0 ; i < run.size()-2 ; ++i) {
					ArrayList<CribbageCard> subrun = new ArrayList<CribbageCard>(run.subList(i, run.size()));
					Collections.sort(subrun);

					boolean isrun = true;
					int val = subrun.get(0).getValue();
					for (int cidx = 1; cidx < subrun.size(); ++cidx) {
						CribbageCard c = subrun.get(cidx);
						if (c.getValue() != ++val) {
							isrun = false;
							break;
						}
					}
					if(isrun) {
						score = subrun.size();
						scorecards  = (CribbageCard[]) subrun.toArray(new CribbageCard[] {});
						break;
					}
				}
				if(score >= 3) {
					result.add(new CribbageScore("Run of "+score, score, scorecards));
				}
			}
			return result;
		}

		private List<CribbageScore> getHandScore(CribbageGameData game, PlayerState player) {
			return getHandScore(game, player, false);
		}
		private List<CribbageScore> getHandScore(CribbageGameData game, PlayerState player, boolean isCrib) {
			
			List<Integer> cards;
			if(isCrib) {
				cards = getCards(game.gameID, CRIB_ROW);
			} else {
				cards = getPlayerHand(game.gameID, player);
			}
			
			// make a copy to modify.
			cards = new ArrayList<Integer>(cards);
			
			List<Integer> play = getCards(game.gameID, PLAY_ROW);

			// insert the cut card at the beginning of the hand (card0)
			cards.add(0, play.get(0));
			
			return getHandScore(cards, isCrib);
		}

		private static List<CribbageScore> getHandScore(List<Integer> cards,
				boolean isCrib) {
			List<CribbageScore> result = new ArrayList<CribbageScore>();
			
			CribbageCard cut = getCard(cards.get(0));
			CribbageCard card1 = getCard(cards.get(1));
			CribbageCard card2 = getCard(cards.get(2));
			CribbageCard card3 = getCard(cards.get(3));
			CribbageCard card4 = getCard(cards.get(4));

			// a Jack that matches the cut card's suit gives 1 point (if Nibs wasn't scored)
			if(cut.getValue() != Card.JACK) {
				for(int i = 1 ; i < cards.size() ; ++i) {
					Card c = getCard(cards.get(i));
					if(c.getValue() == Card.JACK && c.getSuit() == cut.getSuit()) {
						result.add(new CribbageScore("One for His Nob", 1, new Card[] {c}));
						break;
					}
				}
			}
			
			// 2 points for counts of 15, any combination.
			int fifteens = 0;
			for(int setsize = 5 ; setsize >= 2 ; --setsize) {
				List<List<Integer>> cardsets = getSubsets(cards, setsize);
				
				for(List<Integer> cardset : cardsets) {
					if(getCount(cardset) == 15) {
						++fifteens;
					}
				}
			}
			if(fifteens == 1) {
				result.add(new CribbageScore("Fifteen-Two", 2));
			} else if(fifteens > 1) {
				result.add(new CribbageScore("Fifteen-Two ("+fifteens+" ways)", 2*fifteens));
			}
			
			
			// check "of a kinds"
			result.addAll(getPairScores(cards));

			// check for a run of size 5, 4, or 3 (can only be 1 size in a hand of size 5)
			result.addAll(getRunScores(cards));
			
			// a flush
			int suit = card1.getSuit();
			if (suit == card2.getSuit()
					&& suit == card3.getSuit()
					&& suit == card4.getSuit()) {
				if(suit == cut.getSuit()) {
					result.add(new CribbageScore("Flush", 5, new Card[] {cut, card1, card2, card3, card4}));
				} else if(!isCrib) {
					// card flush (not legal in crib hand)
					// only a 4-card flush if cut card is NOT included
					result.add(new CribbageScore("Flush", 4, new Card[] {card1, card2, card3, card4}));
				}
			}
			
			return result;
		}

		private static List<CribbageScore> getRunScores(List<Integer> cardsi) {
			List<Card> cards = new ArrayList<Card>();
			for(Integer c : cardsi) {
				cards.add(getCard(c));
			}
			
			List<CribbageScore> result = new ArrayList<CribbageScore>();

			for(int setsize = cards.size() ; setsize >= 3 ; --setsize) {
				List<List<Card>> cardsets = getSubsets(cards, setsize);

				int score = -1;

				for(List<Card> cardset : cardsets) {
					// make a copy to sort
					ArrayList<Card> runset = new ArrayList<Card>(cardset);
					Collections.sort(runset);
					boolean isrun = true;
//					int val = getCard(runset.get(0)).getValue();
					int val = runset.get(0).getValue();
					for (int cidx = 1; cidx < runset.size(); ++cidx) {
//						CribbageCard c = getCard(runset.get(cidx));
						Card c = runset.get(cidx);
						if (c.getValue() != ++val) {
							isrun = false;
							break;
						}
					}
					if(isrun) {
						score = cardset.size();
						result.add(new CribbageScore("Run of "+score, score, runset.toArray(new Card[0])));
						// don't break, could have more than one run using a pair or whatever
					}
				}

				if(score > 0) {
					// (can only be 1 size of run in a hand of size 5)
					break;
				}
			}
			return result;
		}

		private static List<CribbageScore> getPairScores(List<Integer> cardsi) {
			List<Card> cards = new ArrayList<Card>();
			for(Integer c : cardsi) {
				cards.add(getCard(c));
			}
			
			List<CribbageScore> result = new ArrayList<CribbageScore>();
			
			int samekind = -100;
			List<List<Card>> fourcardsets = getSubsets(cards, 4);
			for(List<Card> cardset : fourcardsets) {
				// check of-a-kinds
//				int kind = getCard(cardset.get(0)).getValue();
				int kind = cardset.get(0).getValue();
				samekind = kind;
				for(int i = 1 ; i < cardset.size() ; ++i) {
//					if(kind != getCard(cardset.get(i)).getValue()) {
					if(kind != cardset.get(i).getValue()) {
						samekind = -100;
						break;
					}
				}
				if(samekind != -100) {
					result .add(new CribbageScore("Double Pair Royal (4-of-a-kind)", 12, cardset.toArray(new Card[0])));
					// only can happen once in a 5 card hand.
					break;
				}
			}
			// if isn't a 4-of-a-kind, there could be one 3-of-a-kind and/or pairs
			if(samekind == -100) {
				List<List<Card>> threecardsets = getSubsets(cards, 3);
				for(List<Card> cardset : threecardsets) {
					// check of-a-kinds
//					int kind = getCard(cardset.get(0)).getValue();
					int kind = cardset.get(0).getValue();
					samekind = kind;
					for(int i = 1 ; i < cardset.size() ; ++i) {
//						if(kind != getCard(cardset.get(i)).getValue()) {
						if(kind != cardset.get(i).getValue()) {
							samekind = -100;
							break;
						}
					}
					if(samekind != -100) {
						result.add(new CribbageScore("Pair Royal (3-of-a-kind)", 6, cardset.toArray(new Card[0])));
						// only can happen once in a 5 card hand.
						break;
					}
				}
				
				// find any pairs that don't match "samekind"'s 3 of a kind
				List<List<Card>> twocardsets = getSubsets(cards, 2);
				for(List<Card> cardset : twocardsets) {
					// check of-a-kinds
//					int kind = getCard(cardset.get(0)).getValue();
					int kind = cardset.get(0).getValue();
//					int kind2 = getCard(cardset.get(1)).getValue();
					int kind2 = cardset.get(1).getValue();
					if(kind == kind2 && kind != samekind) {
						result.add(new CribbageScore("Pair", 2, cardset.toArray(new Card[0])));
					}
				}
			}
			
			return result;
		}
		
		static <T> List<List<T>> getSubsets(List<T> cards, int setsize) {
			
			List<List<T>> result = new ArrayList<List<T>>();

			int indices[] = new int[setsize];
			// initialize indices
			for(int i = 0 ; i < indices.length ; ++i) {
				indices[i] = 0;
			}
			
			result.addAll(getSubsets(indices, 0, cards, setsize));
			
			return result;
		}

		private static <T> Collection<? extends List<T>> getSubsets(int[] indices,
				int i, List<T> cards, int setsize) {

			ArrayList<List<T>> result = new ArrayList<List<T>>();

			if(i >= indices.length) return result;
			if (i > 0) {
				int nextValue = indices[i - 1] + 1;
				if (nextValue > cards.size() - 1) {
					return result;
				} else {
					indices[i] = nextValue;
				}
			} else {
				indices[i] = 0;
			}
			
			while (indices[i] < cards.size()) {
				if(i == setsize-1) {
					ArrayList<T> set = new ArrayList<T>(setsize);
					for(int n = 0; n < setsize ; ++n) {
						set.add(cards.get(indices[n]));
					}
					result.add(set);
				}
				result.addAll(getSubsets(indices, i + 1, cards, setsize));
				++indices[i];
			}
			
			return result;
		}

		private static int getCount(List<Integer> cards) {
			int count = 0;
			for(Integer ref : cards) {
				count += getCard(ref).getCribbageCountValue();
			}
			return count;
		}

		public static ArrayList<CribbageCard> getCards(List<Integer> cards) {
			ArrayList<CribbageCard> result = new ArrayList<CribbageGame.CribbageCard>();
			for (int card : cards) {
				if (card >= 0) {
					CribbageCard ccard = getCard(card);
					result.add(ccard);
				}
			}
			return result;
		}

		public List<Integer> getPlayerHand(int gameID, PlayerState player) {
			int playerRow = getPlayerHandRow(player);

			return getCards(gameID, playerRow);
		}
		private void setPlayerHand(int gameID, PlayerState player, List<Integer> cards) {
			int playerRow = getPlayerHandRow(player);

			 setCards(gameID, playerRow, cards);
		}
		
		//NOTE interprets negatives as facedown
		public CardStack getCardStack(int gameID, int cardRow) {
			return getBoardState(gameID).getCards(cardRow);
		}
		//NOTE interprets facedown to negative
		public void setCardStack(int gameID, int cardRow, CardStack stack) {
			getBoardState(gameID).setCards(cardRow, stack);
		}
		
		public List<Integer> getCards(int gameID, int cardRow) {
			List<Integer> result = new ArrayList<Integer>();

			int i = 0;
			while(i < getBoardState(gameID).width) {
				int card = getBoardState(gameID).get(i, cardRow);
				if (card != 0) {
					result.add(card);
				}
				++i;
			}
			return result;
		}
		
		private void setCards(int gameID, int deckRow, List<Integer> deck) {
			for (int i = 0; i < getBoardState(gameID).width; ++i) {
				if (deck != null && i < deck.size()) {
					getBoardState(gameID).set(i, deckRow, deck.get(i));
				} else {
					getBoardState(gameID).set(i, deckRow, 0);
				}
			}
		}

		public int getPlayerHandRow(PlayerState player) {
			int playerIndex = Arrays.asList(colors).indexOf(player.color);
			int playerRow = PLAYER1_ROW + playerIndex;
			return playerRow;
		}

		
		private void notifyGameChanged(CribbageGameData game, boolean gameStateChanged, boolean gameBoardChanged, boolean playersChanged) {
			if(gameStateChanged || gameBoardChanged) {
				boardUpdated(game.gameID);
			}
			if(playersChanged) {
				playersUpdated(game.gameID);
			}
			
		}

		private void startGame(GameData gamed) {
			CribbageGameData game = (CribbageGameData) gamed;
			setNextDealer(game);
			// only let dealer have turn
			PlayerState dealer = game.players.get(game.dealer);
			for(int i = 0 ; i < game.players.size() ; ++i) {
				game.players.get(i).hasTurn = false;
			}
			dealer.hasTurn = true;
			
			game.gameState = GAMESTATE.ROUND_BEGIN;
			playerChat(game.gameID, dealer, "I'll deal now.");
			
			// notify state, board, and player changes
			notifyGameChanged(game, true, true, true);
		}
		
		private void dealGame(CribbageGameData game) {
			gatherDeck(game);
			shuffleDeck(game);
			dealHands(game);
			// set hasTurn for everybody so they can discard
			for(int i = 0 ; i < game.players.size() ; ++i) {
				game.players.get(i).hasTurn = true;
			}
			// let game go forward.
			game.gameState = GAMESTATE.DISCARDS;
			playerChat(game.gameID, game.players.get(game.dealer), "First, discard 2 for my crib hand");
			
			// notify state, board, and player changes
			notifyGameChanged(game, true, true, true);
		}

		private int setNextDealer(CribbageGameData game) {
			if(game.dealer < 0) {
				game.dealer = 0;
			} else {
				game.dealer  = (game.dealer+1)%game.players.size();
			}
			return game.dealer;
		}

		/**
		 * Shuffle deck 3 times
		 * @param game
		 */
		private void shuffleDeck(CribbageGameData game) {
			CardDeck deck = new CardDeck(getCardStack(game.gameID, DECK_ROW));
			deck.shuffle();
			setCardStack(game.gameID, DECK_ROW, deck.toStack());
		}

		/**
		 * pull cards out of crib, play area, and hands and put them on the deck
		 */
		private void gatherDeck(CribbageGameData game) {
			List<Integer> deck = getCards(game.gameID, DECK_ROW);
			
			List<Integer> crib = getCards(game.gameID, CRIB_ROW);
			deck.addAll(crib);
			crib.clear();
			setCards(game.gameID, CRIB_ROW, crib);
			
			List<Integer> play = getCards(game.gameID, PLAY_ROW);
			// add the turn-up card back to the deck...rest of cards are from player hands
			if(play.size() > 0) {
				deck.add(play.get(0));
			}
			play.clear();
			setCards(game.gameID, PLAY_ROW, play);
			
			for(int i = 0 ; i < game.players.size() ; ++i) {
				int handRow = getPlayerHandRow(game.players.get(i));
				List<Integer> player = getCards(game.gameID, handRow);
				deck.addAll(player);
				player.clear();
				setCards(game.gameID, handRow, player);
			}
			
			setCards(game.gameID, DECK_ROW, deck);
		}

		private void dealHands(CribbageGameData game) {
			//TODO should probably synchronize
			
//			CardDeck deck = new CardDeck(game.boardState.getCards(DECK_ROW));
			
			List<Integer> deck = getCards(game.gameID, DECK_ROW);
			// deal 6 cards...
			for(int c = 0 ; c < DEALT_HAND_SIZE ; ++c) {
				// .. to each player
				for(int i = 0 ; i < game.players.size() ; ++i) {
//					Card card = deck.draw().topCard();
//					game.boardState.setCards(DECK_ROW, deck.toStack());
					
					int card = deck.remove(0); //TODO should probably take from size()-1
					setCards(game.gameID, DECK_ROW, deck);
					
					// turn card face up
					if(card < 0) card = -card;
					
					int handRow = getPlayerHandRow(game.players.get(i));
					List<Integer> player = getCards(game.gameID, handRow);
					player.add(card);
					setCards(game.gameID, handRow, player);
				}
			}
		}

		@Override
		public GameData newGameData(int id) {
			CribbageGame.CribbageGameData gamed = new CribbageGame.CribbageGameData(id);
			gamed.boardState = new BoardState(getBoardWidth(),getBoardHeight());
			// our "no card" value is 0.  negative is a "played" card
			for(int x = 0 ; x < gamed.boardState.width ; ++x) {
				for(int y = 0 ; y < gamed.boardState.height ; ++y) {
					gamed.boardState.set(x, y, 0);
				}
			}
			CardDeck deck = new CardDeck();
			gamed.boardState.setCards(CribbageGame.DECK_ROW, deck.toStack());
			
			gamed.minimumPlayers = 2;
			gamed.maximumPlayers = 2;
//				game.maximumPlayers = 4;
			
			return gamed;
		}

		
		
		
		public static void main(String[] args) {
			List<Integer> testinput1 = Arrays.asList(new Integer[] {5,4,3,2,1});
			List<List<Integer>> answer1 = new ArrayList<List<Integer> >();
			answer1.add(Arrays.asList(new Integer[]{5,4,3}));
			answer1.add(Arrays.asList(new Integer[]{5,4,2}));
			answer1.add(Arrays.asList(new Integer[]{5,4,1}));
			answer1.add(Arrays.asList(new Integer[]{5,3,2}));
			answer1.add(Arrays.asList(new Integer[]{5,3,1}));
			answer1.add(Arrays.asList(new Integer[]{5,2,1}));
			answer1.add(Arrays.asList(new Integer[]{4,3,2}));
			answer1.add(Arrays.asList(new Integer[]{4,3,1}));
			answer1.add(Arrays.asList(new Integer[]{4,2,1}));
			answer1.add(Arrays.asList(new Integer[]{3,2,1}));
			List<List<Integer>> result1 = CribbageGame.getSubsets(testinput1, 3);
			if (result1.equals(answer1)) {
				System.out.println("PASSED subset test");
			} else {
				System.out.println("FAILED subset test");
				System.out.println(result1);
			}

			
			Integer[] testhand1 = new Integer[] {Card.DIAMOND*100+5, Card.SPADE*100+4,Card.HEART*100+5,Card.DIAMOND*100+10,Card.CLUB*100+6};			
			int testscore1 = 16;
			checkScore(testscore1, Arrays.asList(testhand1));

			Integer[] testhand2 = new Integer[] {Card.DIAMOND*100+1, Card.DIAMOND*100+Card.JACK,Card.SPADE*100+3,Card.CLUB*100+3,Card.SPADE*100+2};
			int testscore2 = 13;
			checkScore(testscore2, Arrays.asList(testhand2));
			
			Integer[] testhand3 = new Integer[] {Card.DIAMOND*100+1, Card.HEART*100+3,Card.SPADE*100+3,Card.SPADE*100+4,Card.DIAMOND*100+4};
			int testscore3 = 6;
			checkScore(testscore3, Arrays.asList(testhand3));
			
			// "perfect" hand
			Integer[] testhand4 = new Integer[] {Card.HEART*100+5, Card.HEART*100+Card.JACK,Card.SPADE*100+5,Card.DIAMOND*100+5,Card.CLUB*100+5};
			int testscore4 = 29;
			checkScore(testscore4, Arrays.asList(testhand4));
			
			Integer[] testhand5 = new Integer[] {Card.DIAMOND*100+1, Card.CLUB*100+1,Card.CLUB*100+6,Card.HEART*100+7,Card.SPADE*100+8};
			int testscore5 = 13;
			checkScore(testscore5, Arrays.asList(testhand5));
		
			Integer[] testhand6 = new Integer[] {Card.SPADE*100+1, Card.HEART*100+10,Card.HEART*100+4,Card.SPADE*100+Card.JACK,Card.HEART*100+3};
			int testscore6 = 5;
			checkScore(testscore6, Arrays.asList(testhand6));
			
			// 3 card run, no nobs
			checkPlayScore(3, Arrays.asList(new Integer[] {
					Card.SPADE*100+1, Card.HEART*100+9,Card.HEART*100+Card.KING,Card.HEART*100+Card.QUEEN,Card.SPADE*100+Card.JACK}));
			
			// 15 count
			checkPlayScore(2, Arrays.asList(new Integer[] {
					Card.SPADE*100+1, Card.HEART*100+9,Card.HEART*100+5}));
		
		}

		private static void checkScore(int testscore, List<Integer> testinput) {
			int score = 0;
			List<CribbageScore> result = getHandScore(testinput, false);
			for(CribbageScore s : result) {
				score += s.getValue();
			}
			if(score != testscore) {
				System.out.println("Hand FAILED "+score +"("+testinput+") != "+testscore);
				for(CribbageScore s : result) {
					System.out.println(s+":"+s.getValue());
				}
				
			} else {
				System.out.println("Hand PASSED "+score +"("+testinput+")");
			}				
			
		}
		private static void checkPlayScore(int testscore, List<Integer> testinput) {
			int score = 0;
			List<CribbageScore> result = getPlayScore(testinput, false);
			for(CribbageScore s : result) {
				score += s.getValue();
			}
			if(score != testscore) {
				System.out.println("Play FAILED "+score +"("+testinput+") != "+testscore);
				for(CribbageScore s : result) {
					System.out.println(s+":"+s.getValue());
				}
				
			} else {
				System.out.println("Play PASSED "+score +"("+testinput+")");
			}				
			
		}


}
